from __future__ import annotations

import ast
import asyncio
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterator, List, Optional, cast

from ....jsonrpc2.protocol import rpc_method
from ....utils.async_tools import run_coroutine_in_thread
from ....utils.logging import LoggingDescriptor
from ....utils.uri import Uri
from ...common.lsp_types import (
    DocumentUri,
    Model,
    Position,
    Range,
    TextDocumentIdentifier,
)
from ..utils.async_ast import AsyncVisitor
from .protocol_part import RobotLanguageServerProtocolPart

if TYPE_CHECKING:
    from ..protocol import RobotLanguageServerProtocol


@dataclass(repr=False)
class GetAllTestsParams(Model):
    workspace_folder: str
    paths: Optional[List[str]]
    suites: Optional[List[str]] = None


@dataclass(repr=False)
class GetTestsParams(Model):
    text_document: TextDocumentIdentifier
    base_name: Optional[str]


@dataclass(repr=False)
class GetTestsFromDocumentParams(Model):
    text_document: TextDocumentIdentifier


@dataclass(repr=False)
class TestItem(Model):
    type: str
    id: str
    label: str
    longname: str
    uri: Optional[DocumentUri] = None
    children: Optional[List[TestItem]] = None
    description: Optional[str] = None
    range: Optional[Range] = None
    tags: Optional[List[str]] = None
    error: Optional[str] = None


class FindTestCasesVisitor(AsyncVisitor):
    async def get(self, source: DocumentUri, model: ast.AST, base_name: Optional[str]) -> List[TestItem]:
        self._results: List[TestItem] = []
        self.source = source
        self.path = Uri(source).to_path().resolve()
        self.base_name = base_name
        await self.visit(model)
        return self._results

    async def visit_Section(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.model.blocks import TestCaseSection

        if isinstance(node, TestCaseSection):
            await self.generic_visit(node)

    async def visit_TestCase(self, node: ast.AST) -> None:  # noqa: N802
        from robot.parsing.model.blocks import TestCase
        from robot.parsing.model.statements import Tags

        test_case = cast(TestCase, node)
        longname = f"{self.base_name}.{test_case.name}" if self.base_name else test_case.name
        self._results.append(
            TestItem(
                type="test",
                id=f"{self.path};{longname};{test_case.lineno}",
                longname=longname,
                label=test_case.name,
                uri=self.source,
                range=Range(
                    start=Position(line=test_case.lineno - 1, character=test_case.col_offset),
                    end=Position(
                        line=(test_case.end_lineno if test_case.end_lineno != -1 else test_case.lineno) - 1,
                        character=test_case.end_col_offset if test_case.end_col_offset != -1 else test_case.col_offset,
                    ),
                ),
                tags=[str(tag) for tag in chain(*[tags.values for tags in test_case.body if isinstance(tags, Tags)])],
            )
        )


class DiscoveringProtocolPart(RobotLanguageServerProtocolPart):
    _logger = LoggingDescriptor()

    def __init__(self, parent: RobotLanguageServerProtocol) -> None:
        super().__init__(parent)

    async def _get_tests_from_workspace(
        self, workspace_folder: Path, paths: Optional[List[str]], suites: Optional[List[str]]
    ) -> List[TestItem]:

        from robot.output.logger import LOGGER
        from robot.parsing import get_model
        from robot.parsing.suitestructure import SuiteStructureBuilder
        from robot.running import TestCase, TestSuite
        from robot.running.builder.builders import (
            NoInitFileDirectoryParser,
            RobotParser,
            SuiteStructureParser,
            TestSuiteBuilder,
        )
        from robot.running.builder.testsettings import TestDefaults

        def get_document_text(source: str) -> str:
            if self.parent._loop:
                doc = asyncio.run_coroutine_threadsafe(
                    self.parent.documents.get(Uri.from_path(source).normalized()), self.parent._loop
                ).result()
                if doc is not None and doc.version is not None:
                    return asyncio.run_coroutine_threadsafe(doc.text(), self.parent._loop).result()

            return source

        class MyRobotParser(RobotParser):
            def _get_source(self, source: str) -> Any:
                return get_document_text(source)

            def _build(
                self,
                suite: TestSuite,
                source: str,
                defaults: TestDefaults,
                model: Optional[ast.AST] = None,
                get_model: Callable[..., Any] = get_model,
            ) -> TestSuite:

                from robot.running.builder.transformers import (
                    SettingsBuilder,
                    SuiteBuilder,
                )

                if defaults is None:
                    defaults = TestDefaults()
                if model is None:
                    try:
                        model = get_model(self._get_source(source), data_only=True, curdir=self._get_curdir(source))
                    except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
                        raise
                    except BaseException:
                        pass

                if model is None:
                    return suite

                SettingsBuilder(suite, defaults).visit(model)
                SuiteBuilder(suite, defaults).visit(model)
                suite.rpa = self._get_rpa_mode(model)
                return suite

        class MyRestParser(MyRobotParser):
            def _get_source(self, source: str) -> Any:
                from robot.utils import read_rest_data
                from robot.utils.filereader import FileReader

                with FileReader(source) as reader:
                    return read_rest_data(reader)

        class MySuiteStructureParser(SuiteStructureParser):
            def _get_parsers(self, extensions: List[str], process_curdir: bool) -> RobotParser:
                robot_parser = MyRobotParser(process_curdir)
                rest_parser = MyRestParser(process_curdir)
                parsers = {
                    None: NoInitFileDirectoryParser(),
                    "robot": robot_parser,
                    "rst": rest_parser,
                    "rest": rest_parser,
                }
                for ext in extensions:
                    if ext not in parsers:
                        parsers[ext] = robot_parser
                return parsers

        class MyTestSuiteBuilder(TestSuiteBuilder):
            def _validate_test_counts(self, suite: TestSuite, multisource: bool = False) -> None:
                # we don't need this
                pass

            def build(self, *paths: str) -> TestSuite:
                structure = SuiteStructureBuilder(self.included_extensions, self.included_suites).build(paths)
                parser = MySuiteStructureParser(self.included_extensions, self.rpa, self.process_curdir)
                suite = parser.parse(structure)
                if not self.included_suites and not self.allow_empty_suite:
                    self._validate_test_counts(suite, multisource=len(paths) > 1)
                suite.remove_empty_suites(preserve_direct_children=len(paths) > 1)
                return suite

        def generate(suite: TestSuite) -> TestItem:
            children: List[TestItem] = []

            test: TestCase
            for test in suite.tests:
                children.append(
                    TestItem(
                        type="test",
                        id=f"{Path(test.source).resolve()};{test.longname};{test.lineno}",
                        label=test.name,
                        longname=test.longname,
                        uri=str(Uri.from_path(test.source)) if test.source else None,
                        range=Range(
                            start=Position(line=test.lineno - 1, character=0),
                            end=Position(line=test.lineno - 1, character=0),
                        ),
                        tags=[t for t in test.tags],
                    )
                )

            for s in suite.suites:
                children.append(generate(s))

            return TestItem(
                type="suite",
                id=f"{Path(suite.source).resolve()};{suite.longname}",
                label=suite.name,
                longname=suite.longname,
                uri=str(Uri.from_path(suite.source)) if suite.source else None,
                children=children,
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=0, character=0),
                )
                if suite.source
                else None,
            )

        with LOGGER.cache_only:
            try:
                if paths and len(paths):

                    def normalize_paths(paths: List[str]) -> Iterator[str]:

                        for path in paths:

                            p = Path(path)

                            if not p.is_absolute():
                                p = Path(workspace_folder, p)

                            if p.exists():
                                yield str(p)

                    def nonexisting_paths(paths: List[str]) -> Iterator[str]:

                        for path in paths:

                            p = Path(path)

                            if not p.is_absolute():
                                p = Path(workspace_folder, p)

                            if not p.exists():
                                yield str(p)

                    valid_paths = [i for i in normalize_paths(paths)]
                    suite: Optional[TestSuite] = (
                        MyTestSuiteBuilder(included_suites=suites if suites else None).build(*valid_paths)
                        if valid_paths
                        else None
                    )
                    suite_item = [generate(suite)] if suite else []

                    return [
                        TestItem(
                            type="workspace",
                            id=str(Path.cwd()),
                            label=Path.cwd().name,
                            longname=Path.cwd().name,
                            uri=str(Uri.from_path(Path.cwd())),
                            children=[
                                *suite_item,
                                *[
                                    TestItem(
                                        type="error",
                                        id=f"{i};ERROR",
                                        longname="error",
                                        label=i,
                                        error=f"Parsing '{i}' failed: File or directory to does not exist.",
                                    )
                                    for i in nonexisting_paths(paths)
                                ],
                            ],
                        )
                    ]
                else:
                    return [
                        generate(
                            MyTestSuiteBuilder(included_suites=suites if suites else None).build(str(workspace_folder))
                        )
                    ]
            except (SystemExit, KeyboardInterrupt):
                raise
            except BaseException as e:
                return [
                    TestItem(
                        type="error",
                        id=str(Uri.from_path(Path.cwd())),
                        longname="error",
                        label=Path.cwd().name,
                        error=str(e),
                    )
                ]

    @rpc_method(name="robot/discovering/getTestsFromWorkspace", param_type=GetAllTestsParams)
    @_logger.call
    async def get_tests_from_workspace(
        self,
        workspace_folder: str,
        paths: Optional[List[str]],
        suites: Optional[List[str]],
        *args: Any,
        **kwargs: Any,
    ) -> List[TestItem]:
        return await run_coroutine_in_thread(
            self._get_tests_from_workspace, Uri(workspace_folder).to_path(), paths, suites
        )

    @rpc_method(name="robot/discovering/getTestsFromDocument", param_type=GetTestsParams)
    @_logger.call
    async def get_tests_from_document(
        self, text_document: TextDocumentIdentifier, base_name: Optional[str], *args: Any, **kwargs: Any
    ) -> List[TestItem]:
        async def run() -> List[TestItem]:
            try:
                return await FindTestCasesVisitor().get(
                    text_document.uri,
                    await self.parent.documents_cache.get_model(
                        await self.parent.robot_workspace.get_or_open_document(
                            Uri(text_document.uri).to_path(), language_id="robotframework"
                        )
                    ),
                    base_name,
                )
            except (asyncio.CancelledError, SystemExit, KeyboardInterrupt):
                raise
            except BaseException:
                return []

        return await run_coroutine_in_thread(run)



import abc
import asyncio
import io
import sys
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from types import TracebackType
from typing import (
    BinaryIO,
    Callable,
    Coroutine,
    Generic,
    Literal,
    NamedTuple,
    Optional,
    Type,
    TypeVar,
    cast,
)

from ..utils.logging import LoggingDescriptor
from .protocol import JsonRPCException

__all__ = ["JsonRpcServerMode", "TcpParams", "JsonRPCServer"]

TProtocol = TypeVar("TProtocol", bound=asyncio.Protocol)


class NotSupportedError(Exception):
    pass


class StdOutTransportAdapter(asyncio.Transport):
    def __init__(self, rfile: BinaryIO, wfile: BinaryIO) -> None:
        super().__init__()
        self.rfile = rfile
        self.wfile = wfile

    def close(self) -> None:
        self.rfile.close()
        self.wfile.close()

    def write(self, data: bytes) -> None:
        self.wfile.write(data)
        self.wfile.flush()


class JsonRpcServerMode(Enum):
    STDIO = "stdio"
    TCP = "tcp"
    SOCKET = "socket"
    PIPE = "pipe"


class TcpParams(NamedTuple):
    host: Optional[str] = None
    port: int = 0


class JsonRPCServer(Generic[TProtocol], abc.ABC):
    _logger = LoggingDescriptor()

    def __init__(
        self,
        mode: JsonRpcServerMode = JsonRpcServerMode.STDIO,
        tcp_params: TcpParams = TcpParams(None, 0),
        pipe_name: Optional[str] = None,
    ):
        self.mode = mode
        self.tcp_params = tcp_params
        self.pipe_name = pipe_name

        self._run_func: Optional[Callable[[], None]] = None
        self._serve_func: Optional[Callable[[], Coroutine[None, None, None]]] = None
        self._server: Optional[asyncio.AbstractServer] = None

        self._stdio_stop_event: Optional[asyncio.Event] = None

        self._in_closing = False
        self._closed = False

        self.loop = asyncio.get_event_loop()

    def __del__(self) -> None:
        self.close()

    @_logger.call
    def start(self) -> None:
        if self.mode == JsonRpcServerMode.STDIO:
            self.start_stdio()
        elif self.mode == JsonRpcServerMode.TCP:
            self.loop.run_until_complete(self.start_tcp(self.tcp_params.host, self.tcp_params.port))
        elif self.mode == JsonRpcServerMode.PIPE:
            self.start_pipe(self.pipe_name)
        elif self.mode == JsonRpcServerMode.SOCKET:
            self.start_socket(self.tcp_params.port)
        else:
            raise JsonRPCException(f"Unknown server mode {self.mode}")

    @_logger.call
    async def start_async(self) -> None:
        if self.mode == JsonRpcServerMode.TCP:
            await self.start_tcp(self.tcp_params.host, self.tcp_params.port)
        else:
            raise JsonRPCException(f"Unsupported server mode {self.mode}")

    @_logger.call
    def _close(self) -> None:
        if self._stdio_stop_event is not None:
            self._stdio_stop_event.set()

        if self._server and self._server.is_serving():
            self._server.close()

    @_logger.call
    def close(self) -> None:
        if self._in_closing or self._closed:
            return

        self._in_closing = True
        try:
            self._close()
        finally:
            self._in_closing = False
            self._closed = True

    @_logger.call
    async def close_async(self) -> None:
        if self._in_closing or self._closed:
            return

        self._in_closing = True
        try:
            self._close()
            if self._server is not None:
                await self._server.wait_closed()
        finally:
            self._in_closing = False
            self._closed = True

    async def __aenter__(self) -> "JsonRPCServer[TProtocol]":
        await self.start_async()
        return self

    async def __aexit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        await self.close_async()
        return False

    def __enter__(self) -> "JsonRPCServer[TProtocol]":
        self.start()
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> Literal[False]:
        self.close()
        return False

    @abc.abstractmethod
    def create_protocol(self) -> TProtocol:
        ...

    stdio_executor: Optional[ThreadPoolExecutor] = None

    @_logger.call
    def start_stdio(self) -> None:
        self.mode = JsonRpcServerMode.STDIO

        transport = StdOutTransportAdapter(sys.__stdin__.buffer, sys.__stdout__.buffer)

        protocol = self.create_protocol()

        def run_io_nonblocking() -> None:
            self._stdio_stop_event = asyncio.Event()

            async def aio_readline(rfile: BinaryIO, protocol: asyncio.Protocol) -> None:
                protocol.connection_made(transport)
                stdio_executor = ThreadPoolExecutor(max_workers=1, thread_name_prefix="aio_readline")
                with stdio_executor:
                    while (
                        self._stdio_stop_event is not None and not self._stdio_stop_event.is_set() and not rfile.closed
                    ):
                        data = await self.loop.run_in_executor(
                            stdio_executor, cast(io.BufferedReader, rfile).read1, 1000
                        )
                        protocol.data_received(data)

            self.loop.run_until_complete(aio_readline(transport.rfile, protocol))

        self._run_func = run_io_nonblocking

    @_logger.call
    async def start_tcp(self, host: Optional[str] = None, port: int = 0) -> None:
        self.mode = JsonRpcServerMode.TCP

        self._server = await self.loop.create_server(self.create_protocol, host, port, reuse_address=True)

        self._serve_func = self._server.serve_forever
        self._run_func = self.loop.run_forever

    @_logger.call
    def start_pipe(self, pipe_name: Optional[str]) -> None:
        from typing import TYPE_CHECKING

        if TYPE_CHECKING:
            from asyncio.streams import StreamReaderProtocol
            from asyncio.windows_events import ProactorEventLoop

        if pipe_name is None:
            raise ValueError("pipe name missing.")

        self.mode = JsonRpcServerMode.PIPE

        try:
            #  check if we are on windows and using the ProactorEventLoop, to use the undocumented
            #  create_pipe_connection method
            if sys.platform == "win32" and hasattr(self.loop, "create_pipe_connection"):
                self.loop.run_until_complete(
                    cast("ProactorEventLoop", self.loop).create_pipe_connection(
                        lambda: cast("StreamReaderProtocol", self.create_protocol()), pipe_name
                    ),
                )
            else:
                self.loop.run_until_complete(self.loop.create_unix_connection(self.create_protocol, pipe_name))
        except NotImplementedError:
            raise NotSupportedError("Pipe transport is not supported on this platform.")

        self._run_func = self.loop.run_forever

    @_logger.call
    def start_socket(self, port: int) -> None:
        self.mode = JsonRpcServerMode.SOCKET

        self.loop.run_until_complete(self.loop.create_connection(self.create_protocol, port=port))

        self._run_func = self.loop.run_forever

    @_logger.call
    def run(self) -> None:
        if self._run_func is None:
            self._logger.warning("server is not started.")
            return
        self._run_func()

    @_logger.call
    async def serve(self) -> None:
        if self._serve_func is None:
            self._logger.warning("server is not started.")
            return
        await self._serve_func()



from typing import Any, Dict, Optional

BUILTIN_VARIABLES = [
    "${CURDIR}",
    "${EMPTY}",
    "${TEMPDIR}",
    "${EXECDIR}",
    "${/}",
    "${:}",
    "${\\n}",
    "${SPACE}",
    "${True}",
    "${False}",
    "${None}",
    "${null}",
    "${TEST NAME}",
    "@{TEST TAGS}",
    "${TEST DOCUMENTATION}",
    "${TEST STATUS}",
    "${TEST MESSAGE}",
    "${PREV TEST NAME}",
    "${PREV TEST STATUS}",
    "${PREV TEST MESSAGE}",
    "${SUITE NAME}",
    "${SUITE SOURCE}",
    "${SUITE DOCUMENTATION}",
    "&{SUITE METADATA}",
    "${SUITE STATUS}",
    "${SUITE MESSAGE}",
    "${KEYWORD STATUS}",
    "${KEYWORD MESSAGE}",
    "${LOG LEVEL}",
    "${OUTPUT FILE}",
    "${LOG FILE}",
    "${REPORT FILE}",
    "${DEBUG FILE}",
    "${OUTPUT DIR}",
]


def replace_string(
    value: str,
    working_dir: str = ".",
    base_dir: str = ".",
    command_line_variables: Optional[Dict[str, Optional[Any]]] = None,
    variables: Optional[Dict[str, Optional[Any]]] = None,
    ignore_errors: bool = False,
) -> str:
    from robot.variables import Variables

    from ._variables import RobotCodeVariableReplacer

    vars = Variables()
    vars.store.add("TEST_VAR", "hei there", decorated=False)

    replacer = RobotCodeVariableReplacer(vars.store)
    return str(replacer.replace_string(value, ignore_errors=ignore_errors))
